/*
  Rui Santos
  Complete project details at https://RandomNerdTutorials.com/esp-now-two-way-communication-esp32/
  
  Simplified for demo purposes by R Hart 5/18/20.
  
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files.
  
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
*/

//these are the used libraries
#include <esp_now.h>
#include <WiFi.h>
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27, 20, 4);
#include <Preferences.h>

//GPIO pin definitions
#define yaxis  34 //right joystick, left-right
#define xaxis  35 //right joystick, front-back
#define rbtn   14 //pushbutton of right joystick
#define lbtn   27 //pushbutton of left joystick
#define yaxis2 32 //left joystick, left-right
#define xaxis2 33 //left joystick, front-back
#define vsense 39 //battery voltage of RC (voltage divider)
#define lsw    25 //left switch
#define rsw    26 //right switch
#define encbtn 23 //pushbutton of rotary encoder
#define poti   36 //potentiometer
#define clk    18 //clk input from rotary encoder
#define dt     19 //dt input from rotary encoder

//all the variables. Some could be defined as "bool" instead of "int" to save memory.
int dir            = 0;
int spdL           = 0;
int spdR           = 0;
int yval           = 0;
int xval           = 0;
int speedmax       = 127;
int speedhigh      = 190;
int speedlow       = 100;
int vcrw           = 0;
float voltage      = 0;
int runvar         = 0;
int vbat           = 0;
float RCvol        = 0;
int lswv           = 0;
int rswv           = 0;
int potival        = 0;
int rbtnv          = 1;
int lbtnv          = 1;
int encbtnv        = 1;
int program        = 2;
int yaxis2v        = 0;
int xaxis2v        = 0;
int time0          = 0;
int time1          = 1;
int encdebounce0   = 0;
int encdebounce1   = 0;
int lbtndebounce0  = 0;
int lbtndebounce1  = 0;
int rbtndebounce0  = 0;
int rbtndebounce1  = 0;
int clkdebounce0   = 0;
int clkdebounce1   = 0;
int dtdebounce0    = 0;
int dtdebounce1    = 0;
int timer          = 0;
int clkv           = 1;
int dtv            = 1;
int rot            = 0;
int encchk         = 1;
int lcdprogram     = 0;
int ledv           = 200;
int rotdir         = 1;
int adj            = 0;
int controlmode    = 1;
int runvar1        = 0;
int signalquality  = 0;
int dnsv           = 0;
int counter        = 0;
int drivemode      = 0;
int defaultheight  = 0;
int potivalold     = 0;
int rbr            = 0;
int rbg            = 0;
int rbb            = 130;
int lighton        = 0;
int batcounter     = 0;
int error          = 0;
int rxbatwarn      = 0;
int txbatwarn      = 0;
int batwarn        = 0;
int voltagecounter = 0;
int bullets        = 0;
int force          = 4;
int shoot          = 0;
int wpncounter     = 0;
int modeadj        = 0;

//definition of some non-standard characters for the LCD display
byte chr1[8] =  { 
                  B00000,
                  B00100,
                  B01110,
                  B11111,
                  B11111,
                  B01110,
                  B00100,
                  B00000
              };
              
byte chr2[8] =  { 
                  B11111,
                  B10001,
                  B10001,
                  B10001,
                  B10001,
                  B10001,
                  B10001,
                  B11111
              };  

byte chr3[8] =  { 
                  B11111,
                  B10001,
                  B10001,
                  B10001,
                  B10001,
                  B11111,
                  B11111,
                  B11111
              };  

byte chr4[8] =  { 
                  B11111,
                  B10001,
                  B10001,
                  B11111,
                  B11111,
                  B11111,
                  B11111,
                  B11111
              }; 

byte chr5[8] =  { 
                  B11111,
                  B11111,
                  B11111,
                  B11111,
                  B11111,
                  B11111,
                  B11111,
                  B11111
              }; 

byte chr6[8] =  { 
                  B11111,
                  B10001,
                  B01010,
                  B00100,
                  B01010,
                  B10001,
                  B00000,
                  B11111
              };               
                            
//ESP-Now requires the MAC address of the other board to connect successfully. Thus, you need to enter the MAC address of your ESP pairs below.
uint8_t broadcastAddress[] =  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};  //MAC address of Primary ESP32

//Definition of the data package that is sent to the other ESP32
typedef struct send_message {
  int spdL;
  int spdR;
  int hv;
  int R;
  int G;
  int B;
  int wpn;
  int ls;
  int drv;
} send_message;
 
// Create a structured object
send_message outgoing;

//The data package for receiving data needs to be defined, too
typedef struct receive_message {
  int voltage;
  int errorcode;
  int drumpos;
} receive_message;
 
// Create a structured object
receive_message incoming;

// Callback when data is sent
void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  if (status == ESP_NOW_SEND_SUCCESS&&signalquality<19){ //this feedback is used to detect the signal quality (on a scale from 0 to 19)
    signalquality = signalquality +1;
  }else if(signalquality>0){
    signalquality = signalquality -1;
  }
}

// Callback when data is received
void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  memcpy(&incoming, incomingData, sizeof(incoming));
}

esp_now_peer_info_t peerInfo; //be aware of this line! It has to be below the other ESP-Now related stuff in the sketch, or it won't work.

//the rotary encoder is detected using an interrupt, with a debounce time interval of 500 ms
//note that the interrupts need the "IRAM_ATTR" parameter to be executed in the RAM of the µC
void IRAM_ATTR encint(){
  encdebounce1 = millis();
  if (encdebounce1>encdebounce0+500){
    encbtnv = 0;    
  }
  encdebounce0 = encdebounce1;
}

//the left button uses an interrupt with several functions: It is used to reset the main body tilt, to shoot, and to reload (depending on the other settings)
void IRAM_ATTR lbtnint(){
  lbtndebounce1 = millis();
  if (lbtndebounce1>lbtndebounce0+500){
    if (lswv==0){
      if(lcdprogram==8&&adj==2){
        shoot = 2; 
      }else{ 
        lbtnv = 0;  
      }       
    }else{
      shoot = 1;
    }
  }
  lbtndebounce0 = lbtndebounce1;
}

//the right button. It is used to switch the headlights on and off.
void IRAM_ATTR rbtnint(){
  rbtndebounce1 = millis();
  if (rbtndebounce1>rbtndebounce0+500){
    rbtnv = 0;    
  }
  rbtndebounce0 = rbtndebounce1;
}

//the rotation of the rotary encoder is detected with this interrupt (shorter debounce interval of 100 ms)
void IRAM_ATTR clkint(){
  clkdebounce1 = millis();
  if (clkdebounce1>clkdebounce0+100){
    clkv = 0;
    dtv = digitalRead(dt);  //the µC checks the rotation direction by triggering clk using the interrupt and then reading dt within the interrupt, which results either in 0/0 or 0/1 - both directions covered  
  }
  clkdebounce0 = clkdebounce1;
}

//the preferences library enables you to store values of variables within the flash so that you can retrieve them during the next startup of the µC. Used here to store customizable settings of the remote control
Preferences preferences;
 
void setup() {
  preferences.begin("Defaults", false);
  program       = preferences.getInt("RGBstatus", 0);
  ledv          = preferences.getInt("LEDstatus", 3);
  speedhigh     = preferences.getInt("SPDH", 150);
  speedlow      = preferences.getInt("SPDL", 80);
  controlmode   = preferences.getInt("Control", 0);
  drivemode     = preferences.getInt("Drive", 0);
  defaultheight = preferences.getInt("Height",30);
  bullets       = preferences.getInt("Ammo", 0);
  force         = preferences.getInt("Force",4);

  pinMode(lsw, INPUT_PULLUP);
  pinMode(rsw, INPUT_PULLUP);
  pinMode(rbtn, INPUT_PULLUP);
  pinMode(lbtn, INPUT_PULLUP);
  pinMode(encbtn, INPUT_PULLUP);
  pinMode(clk, INPUT);
  pinMode(dt, INPUT);

  //initial definition of values to be sent/received
  outgoing.R         = 0;
  outgoing.G         = 0;
  outgoing.B         = 0;
  outgoing.wpn       = 0;
  outgoing.spdL      = 0;
  outgoing.spdR      = 0;
  incoming.errorcode = 0;
  incoming.drumpos   = 16;
  
  potival = analogRead(poti);
  for (int i=0; i<19; i++){ //the desired height value is averaged to reduce noise
    potival = potival+analogRead(poti);
  }
  potival = potival/20;
  if (rswv>0){
    potival = map(potival,0,4095,40,55); //the maximum height of the tank is reduced in the fast drive mode
  }else{
    potival = map(potival,0,4095,15,55); //this is the full height range: 15 (highest) to 55 (lowest)
  }
  outgoing.hv = potival;
  potivalold  = potival;

  lswv = digitalRead(lsw);
   
  lcd.init();
  lcd.createChar(1, chr1); //the custom characters for the LCD are provided to the library
  lcd.createChar(2, chr2);
  lcd.createChar(3, chr3);
  lcd.createChar(4, chr4);
  lcd.createChar(5, chr5);
  lcd.createChar(6, chr6);
  lcd.backlight();
  lcd.setCursor(0,0);
  lcd.print("Initializing"); //startup sequence of the remote control
  delay(250);
  lcd.print(".");
  delay(250);
  lcd.print(".");
  delay(250);
  lcd.print(".");
  delay(250);
  while (lswv>0){ //Do not enter main loop if the weapon is switched to "active" -> wait here until it is disabled
    lcd.setCursor(2,2);
    lcd.print("Disable weapon");
    lcd.setCursor(2,3);
    lcd.print("during startup!");
    delay(500);
    lswv = digitalRead(lsw);
  }
  lcd.setCursor(2,2);
  lcd.print("              ");
  lcd.setCursor(2,3);
  lcd.print("               ");
  lcd.setCursor(0,1);
  lcd.print("Battery: ");
  for (int i=0; i<199; i++){
    vbat = vbat + analogRead(vsense);
  }
  RCvol = float(vbat/200)/4095; //voltage readout of the battery of the remote control (mean of 200 measurements)
  vbat  = 0;
  RCvol = (992.2*RCvol*RCvol*RCvol-2658*RCvol*RCvol+2386*RCvol-709.4)-0.35; //correction function to obtain +-0.1V accuracy between 6.3 and 8.5V (1.05 to 1.42V per cell)
  //the correction is individual for each ESP32 - you need to do that with a multimeter and a curve fitting software
  //it would be much better if the ADCs of the ESP32 were more linear...
    if (RCvol<6.8){
      error = 1; //6.8 V (a bit more than 1.1 V per cell) triggers the low battery error: RC Operation is terminated at this threshold
    }
  lcd.setCursor(9,1);
  lcd.print(RCvol,1);
  lcd.print(" V");
  delay(1000);

  
  //The serial monitor is needed for debugging the µC
  Serial.begin(115200);
  
  //The following code lines are needed for ESP-Now
  // Set device as a Wi-Fi Station
  WiFi.mode(WIFI_STA);

  // Init ESP-NOW
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);
  
  // Register peer
  //esp_now_peer_info_t peerInfo;
  memcpy(peerInfo.peer_addr, broadcastAddress, 6);
  peerInfo.channel = 0;  
  peerInfo.encrypt = false;
  
  // Add peer        
  if (esp_now_add_peer(&peerInfo) != ESP_OK){
    Serial.println("Failed to add peer");
    return;
  }

  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
  //End of the ESP-Now related code

  time0 = millis();
  attachInterrupt(encbtn,encint,FALLING); //All buttons use pullup-resistors (wired to GND) -> interrupts are triggered with a falling signal
  attachInterrupt(lbtn,lbtnint,FALLING);
  attachInterrupt(rbtn,rbtnint,FALLING);
  attachInterrupt(clk,clkint,FALLING);

  if (error==0){
    lcd.setCursor(0,2);
    lcd.print("Waiting for Cxn:");
    lcd.setCursor(0,3);
    outgoing.hv = defaultheight; //defaultheight specifies whether the tank auto-adjusts its tilt or if tilt adjust is done manually.
    //it is only sent when the mode is changed. Therefore, it is sent once during startup to make sure the current setting is sent to the tank.
    while (signalquality<10&&rot==0){ //at least 10 out of 20 data packages need to be received so that the connection is considered as established
      if (rbtnv==0&&lbtnv==0){
        rot=1; //the remote control does not enter the main menu until the connection to the tank is established - or if the two buttons are pressed
      }    
      esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &outgoing, sizeof(outgoing)); 
      delay(50);
      runvar=runvar+1; //the next lines show the running symbol line as an indicator for "I'm not stuck forever, I'm doing something!"
      if (runvar>2){
        runvar=0;
        if (runvar1<19){
          if (adj==0){
            lcd.write(1);
          }else{
            lcd.print(" ");
          }     
          runvar1 = runvar1+1;
        }else{
          runvar1=0;
          if(adj==0){
            adj = 1;
            lcd.write(1);
          }else{
            adj = 0;
            lcd.print(" ");
          }
          lcd.setCursor(0,3);      
        }
      }  
    }
    lcd.setCursor(0,3);
    if (signalquality>5){
      lcd.print("Connection ready!   ");
    }else{
      lcd.print("Search aborted!     ");
    }  
    delay(2000);
  }
  lcd.clear();
  runvar  = 0;
  adj     = 0;
  runvar1 = 0;
  rot     = 0;
  rbtnv   = 1;
  lbtnv   = 1;
}
 
void loop() {

  if (incoming.errorcode>0){ //the error codes are related to low battery or motor train issues (gearbox broken or motor stalling)
    error = incoming.errorcode;
  }

  if (incoming.drumpos<17){ //this is the ammunition count (16 bullets = full)
    bullets = 16-incoming.drumpos;
    preferences.putInt("Ammo", bullets);
  }else{
    bullets = 17; //a value of 17 corresponds to the weapon error (=jammed during reloading)
  }
  
  if (rbtnv==0){ //switching the headlights on and off
    rbtnv = 1;
    if (lighton==0){
      lighton = 1;
    }else{
      lighton = 0;
    }
  }
  if (lighton==1){
    outgoing.ls = ledv*63; //simple transmission of an 8bit duty cycle (0-255) with 4 intensity values: 63 - 126 - 189 - 252
  }else{
    outgoing.ls = 0;
  }  
  
  rswv = digitalRead(rsw); //change between fast and slow drive mode
  if (rswv>0){
    speedmax = speedhigh;
  }else{
    speedmax = speedlow;
  }
  
  lswv = digitalRead(lsw); //(de)activation of the weapon
  if (lswv==1){
    if (modeadj==0){ //activating the weapon switches to manual height adjustment (defaultheight=111) and controlling the tilt with the left joystick (controlmode=0)
      if (defaultheight==110&&controlmode==1){
        modeadj       = 1;
        defaultheight = 111;
        controlmode   = 0;
      }else if (defaultheight==110&&controlmode==0){
        modeadj       = 2;
        defaultheight = 111;
      }else if (defaultheight==111&&controlmode==1){
        modeadj       = 3;
        controlmode   = 0;
      }
      if (modeadj>0){
        outgoing.hv = defaultheight;
        for (int i=0; i<5; i++){ //there is no feedback from the tank if the mode change has been received. Thus, it is sent 5x so that it is very likely received
          esp_now_send(broadcastAddress, (uint8_t *) &outgoing, sizeof(outgoing)); 
          delay(25);        
        }
      }
    }
  }else{
    if (modeadj>0){ //if the controls were changed by switching the weapon on, they will be set back to their initial state when deactivating the weapon
      if (modeadj==1){
        defaultheight = 110;
        controlmode   = 1;
        modeadj       = 0;
      }else if (modeadj==2){
        defaultheight = 110;
        modeadj       = 0;
      }else if (modeadj==3){
        controlmode = 1;
        modeadj     = 0;
      }
      outgoing.hv=defaultheight;
      for (int i=0; i<5; i++){
        esp_now_send(broadcastAddress, (uint8_t *) &outgoing, sizeof(outgoing)); 
        delay(25);        
      }    
    }
  }
  if (lswv==1||shoot==2){ //transmitting weapon force, shoot, and reload commands is done with the outgoing.wpn variable
    if (shoot>0){
      if (shoot==1){ //shoot=1 when pressing the left button while the weapon is activated -> fire
        outgoing.wpn = 5;       
      }else if (shoot==2){ //shoot=2 when pressing the left button in the weapon reload menu of the RC -> reloading completed
        outgoing.wpn = 6; 
      }
      wpncounter = wpncounter + 1;
      if (wpncounter==50){ //this additional counter ensures that the fire/reload command is saved in outgoing.wpn long enough for the tank to receive it...
        shoot      = 0;  
        wpncounter = 0;      
      }
    }else{
      outgoing.wpn = force; //...because the default data that is sent is the weapon force between 1 and 4...   
    }  
  }else{
    outgoing.wpn = 0; //...or 0 when the weapon is deactivated
  }
  
  vbat = vbat + analogRead(vsense);
  if (incoming.voltage>10){ //the voltage readout of the tank is converted into a voltage within the RC: It receives an analog value; too small values are ignored
    vcrw           = vcrw + incoming.voltage;
    voltagecounter = voltagecounter+1;
  }
  if (runvar>199){
    runvar = 0;
    RCvol  = float(vbat/200)/4095;
    RCvol  = (992.2*RCvol*RCvol*RCvol-2658*RCvol*RCvol+2386*RCvol-709.4)-0.35; //correction function to obtain +-0.1V accuracy between 6.3 and 8.5V (1.05 to 1.42V per cell)
    if (RCvol<7.0){ //low battery warning level: below 7.0V
      txbatwarn = 1;
    }else if (RCvol>7.2){ //low battery warning dissappears at voltages above 7.2V
      txbatwarn = 0;
    }
    if (RCvol<6.8&&error==0){ //the threshold for low battery error is triggered below 6.8V when this value was detected 100 times
      batcounter = batcounter +1;
    }else if (batcounter>0&&error==0){
      batcounter = batcounter -1;
    }
    if (batcounter>100&&error==0){
      error = 1;
      lcd.clear();
    }
    
    if (voltagecounter>0){ //the voltage of the tank battery is only calculated when there was an input (ignored when there is no connection)
      voltage = float(vcrw/voltagecounter)/4095; //the value is divided by the variable that saves the number of received inputs (important! otherwise, readings of 0 skew the result)
      voltage = 18.3682*voltage+0.923244; //the individual correction function for the ADC of the primary ESP32 within the tank - individual measurement&determination of the correction function required!
    }else{
      voltage = 0;
    }

    if (voltage<11.2&&voltagecounter>190){ //low battery warning for the tank battery at less than 11.2 V. Low battery error is determined by the primary ESP32 of the tank
      rxbatwarn = 1;
    }else if (voltage>11.4){
      rxbatwarn = 0;
    }
    vbat           = 0;
    vcrw           = 0;
    voltagecounter = 0;
  }
  runvar = runvar +1;

  if (incoming.errorcode>1){
    error = incoming.errorcode;
    lcd.clear();
  }

  if (error==0){ //error messages overwrite the standard loop, and they can only be removed by restarting both tank and remote control
    if (counter==0){ //the standard loop does not call all functions at every iteration but calls different functions every iteration to reduce the time spent per loop iteration
      heightAdjust();    
    }else if(counter==1){
      RGBupdate(); 
    }else if(counter==2){
      switch (lcdprogram){ //the different pages of the LCD screen are called as different functions depending on the current selection (saved in lcdprogram)
        case 0:
          updateLCD(); //standard screen
          break;
        case 1:
          LCDenc(); //main menu screen
          break;
        case 2:
          Lightenc(); //headlight adjust screen
          break;
        case 3:
          Speedenc(); //speed adjust screen
          break;  
        case 4:
          Controlenc(); //control mode adjust screen (joystick functions)
          break;
        case 5:
          Driveenc(); //drive mode adjust screen (direct or interpolated)
          break; 
        case 6:
          Heightenc(); //height mode adjust screen (manual or automatic tilt)
          break;
        case 7:
          RGBenc(); //RGB strip mode adjust screen (color selection or off)
          break; 
        case 8:
          Weaponenc(); //weapon settings screen
          break;        
      }
    }    
  }else{
    ErrorScreen();
  }

  time1 = millis();
  if (time1>time0+25){    
    if (lcdprogram!=6){ //do not allow reset during height adjust screen
      HeightReset(); //reset of height needs to run more often than height adjust to account for the limited trigger time of the interrupt
    }else{
      outgoing.hv=defaultheight;
    }
    counter = counter +1;
    if (counter==3){
      if (rbr==0&&rbb>0){ //these variables correspond to RGB values for the rainbow mode of the RGB LED strip: rb equals rainbow, and the individual colors r, b, and g
        rbb = rbb-2; //the change of each value per iteration is timed based on the 25 ms delay and the counter variable
        rbg = rbg+2; //in total, this code snippet results in a rainbow color change from blue to green to red to blue and so on
      }else if(rbb==0&&rbg>0){
        rbg = rbg-2;
        rbr = rbr+2;
      }else if(rbg==0&&rbr>0){
        rbr = rbr-2;
        rbb = rbb+2;
      }      
      counter = 0;
      //updateSerial(); //optional - used for debugging
    }
    if (controlmode==0){ //speed inputs are read with different functions depending on slow or fast drive modes
      speedchk0();
    }else{
      speedchk1();
    } 
    Encoderchk();
    outgoing.drv = drivemode; //the drivemode (direct or interpolated) is calculated within the tank to avoid abrupt speed changes in interpolated mode due to connection issues
    if (error>0){ //safety feature: speed, headlight, and weapon are zero/off when there is an error
      outgoing.spdL = 0;
      outgoing.spdR = 0;
      outgoing.ls   = 0;
      outgoing.wpn  = 0;      
    }
    esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &outgoing, sizeof(outgoing)); 

    time0 = millis();
  }

}

//this function is responsible for the weapon adjust screen
void Weaponenc(){
  lcd.setCursor(0,0);
  lcd.print("Weapon control");
  lcd.setCursor(1,1);
  lcd.print("Force:");
  lcd.setCursor(8,1);
  lcd.print(force);
  lcd.setCursor(10,1);
  lcd.print("(1-4)");

  lcd.setCursor(1,2);
  lcd.print("Reload:");
  lcd.setCursor(9,2);
  if (bullets==16){
    lcd.print("ammo full  ");   
  }else{
    if (lswv==1){ //reloading is prohibited while the weapon is active - indicate this feature as a warning
      lcd.print("wpn active!"); 
    }else{
      lcd.print("left button"); 
    }
  }
  lcd.setCursor(10,3);
  lcd.print("Ammo: ");
  if (bullets==17){ //this value (17) is the sign for a jammed weapon
    lcd.print("XX");
  }else if(bullets>9){
    lcd.print(bullets);
  }else{
    lcd.print("0");
    lcd.print(bullets);
  }
  
  lcd.setCursor(1,3);
  lcd.print("Back");

  if (adj==0){ //within this menu, the adj variable corresponds to the current level (selecting, adjusting force, or reloading)
    for (int i=0; i<3; i++){ //within the selecting subset, the rot variable corresponds to where the cursor is
      lcd.setCursor(0,i+1);
      if (rot==i){
        lcd.write(1);
      }else{
        lcd.print(" ");
      }
      lcd.setCursor(7,1);
      lcd.print(" ");
      lcd.setCursor(8,2);
      lcd.print(" ");
    }
  }else{ //when adj>0, the adj variable indicates whether the force is adjusted or if reloading is selected (indicated by the cursor)
    if (adj==1){
      lcd.setCursor(7,1);
      lcd.write(1);
      lcd.setCursor(8,2);
      lcd.print(" ");
    }else{
      lcd.setCursor(8,2);
      lcd.write(1);
      lcd.setCursor(7,1);
      lcd.print(" "); 
    }
  }  
}

//the error screen displays the error code with a description of the error (is only entered when there is an error)
void ErrorScreen(){
  lcd.setCursor(4,0);
  lcd.print("Error ");
  switch (error){
    case 1:
      lcd.print("01:");
      break;
    case 2:
      lcd.print("02:");
      break;
    case 3:
      lcd.print("03:");
      break;
    case 4:
      lcd.print("04:");
      break;            
  }
  lcd.setCursor(0,2);
  switch (error){
    case 1:
      lcd.print("Transmitter battery");
      lcd.setCursor(0,3);
      lcd.print("low");
      break;
    case 2:
      lcd.print("Receiver battery low");
      break; 
    case 3:
      lcd.print("Motor error left");
      break;
    case 4:
      lcd.print("Motor error right");
      break;              
  }
  delay(2000);
}

//the RGB LED adjust screen
void RGBenc(){
  lcd.setCursor(0,0);
  lcd.print("Running light mode:");
  lcd.setCursor(0,1);
  lcd.print("    ");
  lcd.setCursor(4,1);
  switch (program){
    case 0:
      lcd.print("Off    ");
      break;
    case 1:
      lcd.print("Blue   ");
      break;  
    case 2:
      lcd.print("Cyan   ");
      break;
    case 3:
      lcd.print("Green  ");
      break;      
    case 4:
      lcd.print("Yellow ");
      break;
    case 5:
      lcd.print("Red    ");
      break;
    case 6:
      lcd.print("Violet ");
      break;      
    case 7:
      lcd.print("White  ");
      break;
    case 8:
      lcd.print("Rainbow");
      break;    
  }
  lcd.setCursor(0,3);
  lcd.print("RGB:"); //Not only the color, but also the RGB code of the color is displayed
  lcd.setCursor(5,3);
  lcd.print("[");
  if (outgoing.R<10){
    lcd.print("00");
  }else if (outgoing.R<100){
    lcd.print("0");
  }  
  lcd.print(outgoing.R);
  lcd.print(",");
  if (outgoing.G<10){
    lcd.print("00");
  }else if (outgoing.G<100){
    lcd.print("0");
  }  
  lcd.print(outgoing.G);
  lcd.print(",");
  if (outgoing.B<10){
    lcd.print("00");
  }else if (outgoing.B<100){
    lcd.print("0");
  }   
  lcd.print(outgoing.B);
  lcd.print("]");   
}

//Tilt adjust mode screen
void Heightenc(){
  lcd.setCursor(0,0);
  lcd.print("Tilt adjust mode: ");
  lcd.setCursor(3,1);
  if (defaultheight==110){ //110 = auto; 111 = manual
    lcd.print("Auto  ");  
  }else if (defaultheight==111){
    lcd.print("Manual");
  }
  lcd.setCursor(0,2);
  if (defaultheight==111){ //this is a description of what the joysticks control depending on which mode is selected
    lcd.print("(");
    lcd.write(1);
    lcd.print(") + Left stick    ");
    lcd.setCursor(0,3);
    lcd.print("Reset: left button  ");
  }else if(defaultheight==110){
    lcd.print("Self-balancing      ");
    lcd.setCursor(0,3);
    lcd.print("(only when standing)");
  }
}

//pressing the left button resets the tilt of the tank; this function only triggers the tilt reset when the weapon is off; the command is only processed by the tank if it is set to the manual tilt adjust mode
void HeightReset(){
  if (lbtnv==0){
    timer = timer+1;
    if (timer>9){
      lbtnv = 1;
      timer = 0;
    }   
    outgoing.hv = 109;
  }
}

//this function is responsible for entering the LCD menu, the different screens, and for entering the setting adjustments within the single screens of the menu
void Encoderchk(){
  if (encbtnv==0){
    if (encchk==1){ //start reading encoder rotation: encchk is 1 while being in the main menu and turns to 0 when entering the menu screen
      encchk = 0;
      if (lcdprogram==0){ //go to menu screen
        lcdprogram = 1;
      }
    }else{ //on second encoder butten press (=when being within the menu screen), ...
      if (lcdprogram==1){ 
        if (rot==0){ //...return to standard screen
          encchk     = 1; //encchk=1 means the encoder rotation is not recorded any longer
          rot        = 0; 
          lcdprogram = 0;
        }else if(rot==1){ //...go to LED intensity change screen
          rot        = ledv;
          lcdprogram = 2;
        }else if (rot==2){ //...go to speed adjust screen
          rot        = 0;
          lcdprogram = 3;
        }else if (rot==3){ //...go to control adjust screen
          rot        = controlmode;
          lcdprogram = 4;
        }else if (rot==4){ //...go to drive adjust screen
          rot        = drivemode;
          lcdprogram = 5;
        }else if (rot==5){ //...go to tilt mode screen
          lcdprogram = 6; 
        }else if (rot==6){ //...go to RGB adjust screen
          rot        = program;
          lcdprogram = 7; 
        }else if (rot==7){ //...go to Weapon adjust screen
          rot        = 0;
          lcdprogram = 8;
        }
      }else if(lcdprogram>1&&lcdprogram!=3&&lcdprogram!=8){ //...go back to menu screen when being within one of the adjustment screens (but not speeed or weapon - they have sub-selection possibilities)
        rot        = lcdprogram-1; //let the cursor appear at the point of the screen that you just left
        lcdprogram = 1;                     
      }else if(lcdprogram==3&&adj==0){ //sub-selections within the speed adjustment menu...
        if (rot==0){ //...adjustment of maximum speed in fast mode...
          adj = 1;
          rot = speedhigh;
        }else if (rot==1){ //...adjustment of maximum speed in slow mode...
          adj = 2;
          rot = speedlow;
        }else{ //...return to menu screen
          rot        = lcdprogram-1;
          lcdprogram = 1;         
        }
      }else if ((lcdprogram==3||lcdprogram==8)&&adj>0){ //when having adjusted something within speed or weapon adjustment screen, let the cursor return to the old position
        rot = adj-1;
        adj = 0;
      }else if (lcdprogram==8&&adj==0){ //sub-selections within the weapon adjustment menu...
        if (rot==0){ //...adjustment of the force (=speed of the acceleration wheels)...
          adj = 1;
          rot = force;
        }else if (rot==1){ //...reloading option...
          adj = 2; //reloading is triggered within the interrupt of the left button, so no further variable assignment needed here!
        }else{ //...return to menu screen
          rot        = lcdprogram-1;
          lcdprogram = 1;
        }
      }
    }
    encbtnv = 1;   
    lcd.clear(); //the LCD screen is cleared once the encoder is pressed (=every time a screen page is changed): Helps to avoid remnant characters somewhere on the screen
  }

  if (encchk==0){ //when the encoder rotation is actively monitored (within the menu screens), the rotation direction is checked (0 or 2; 1 means no rotation was observed)
    if (clkv==0){
      if (dtv==1){
        rotdir = 2;
      }else{
        rotdir = 0;  
      }
      clkv = 1;
    }else{
      rotdir = 1;       
    }
    
    switch (lcdprogram){
      case 1: //when being within the menu selection screen, the cursor position can change between all eight menu entries according to whether a rotation (and which direction) was detected
        if (rotdir==2){
          rot = rot+1;
        }else if (rotdir==0){
          rot = rot-1;
        }
        if (rot<0){
          rot = 7;
        }else if(rot>7){
          rot = 0;
        }  
        break;
      case 2: //when being within the light intensity adjustment screen, the encoder rotation determines the light intensity between 1 and 4
        if (rotdir==2&&rot<4){
          rot = rot+1;
        }else if (rotdir==0&&rot>1){
          rot = rot-1;
        }
        ledv = rot;
        preferences.putInt("LEDstatus", ledv); //adjustments of the LED headlights are stored within the flash: will be retrieved when restarting the remote control
        break;
      case 3: //when being within the speed adjustment screen...
        if (adj==0){ //...the encoder rotation determines the cursor position while not having selected a speed level to adjust...
          if (rotdir==2){
            rot = rot+1;
          }else if (rotdir==0){
            rot = rot-1;
          }
          if (rot>2){
            rot = 0;
          }else if (rot<0){
            rot = 2;
          }
        }else{ //...or the maximum speed in slow and fast modes
          if (rotdir==2&&rot<199){ //absolute maximum: speed of 200 (equals a duty cycle of 200 of an 8bit PWM channel (0-255)
            rot = rot+5;
          }else if (rotdir==0&&rot>54){ //absolute minimum: speed of 50
            rot = rot-5;
          }
          if (adj==1){ //adjustment of the maximum speed in the fast mode
            speedhigh = rot;             
            if ((speedhigh<=speedlow+20)&&(speedhigh-25>49)){ //dependency: fast mode has to be at least 25 more than slow mode
              speedlow = speedhigh-25;
            }
            if (speedhigh<75){ //minimum for fast mode: 75
              speedhigh = 75;
              rot       = speedhigh;
            }
          }else{ //adjustment of the maximum speed in the slow mode
            speedlow = rot;            
            if ((speedhigh<=speedlow+20)&&(speedlow+25<201)){ //dependency: slow mode has to be at least 25 less than fast mode
              speedhigh = speedlow+25;
            }
            if (speedlow>175){ //maximum for slow mode: 175
              speedlow = 175;
              rot      = speedlow;
            }
          }
          preferences.putInt("SPDH", speedhigh); //the speed levels are stored as preferences: settings available on next startup
          preferences.putInt("SPDL", speedlow);
        }
        break;
      case 4: //adjustment of the control mode (two joysticks for driving or only one)
        if (rotdir!=1){
          if (controlmode==0){
            controlmode = 1;               
          }else{
            controlmode = 0;
          }
          preferences.putInt("Control", controlmode); //also saved as a preference
        }
        break;
      case 5: //adjustment of direct or interpolated drive mode
        if (rotdir!=1){
          if (drivemode==0){
            drivemode = 1;  
          }else{
            drivemode = 0;
          }
          preferences.putInt("Drive", drivemode); //also saved as a preference
        }
        break;
      case 6: //adjustment of the tilt adjust mode (auto or manual)
        if (rotdir!=1){
          if (defaultheight==111){
            defaultheight = 110;  
          }else{
            defaultheight = 111;
          }
          preferences.putInt("Height", defaultheight); //also saved as a preference
        }
        break;
      case 7: //adjustment of RGB LED strip mode
        if (rotdir>1){ //switching through the nine different color modes
          if (program<8){
            program = program+1;
          }else{
            program = 0;
          }
        }else if (rotdir<1){
          if (program>0){
            program = program-1;
          }else{
            program = 8;
          }
        }
        preferences.putInt("RGBstatus", program); //also saved as a preference
        break;
      case 8: //weapon adjustment: cursor movement
        if (adj==0){
          if (rotdir==2){
            rot = rot+1;
          }else if (rotdir==0){
            rot = rot-1;
          }
          if (rot>2){
            rot = 0;
          }else if (rot<0){
            rot = 2;
          }
        }else{
          if (adj==1){ //weapon force adjustment between one and four
            if (rotdir==2&&rot<4){
              rot = rot+1;
            }else if (rotdir==0&&rot>1){
              rot = rot-1;
            }
            force = rot;
            preferences.putInt("Force", force);
          }else{
            //check left button: pressing = reload
            preferences.putInt("Ammo", bullets); //save bullet count in case it changed
          }
        }          
        break;          
    }
  }  
}

//LCD page for drive mode adjustment
void Driveenc(){
  lcd.setCursor(0,0);
  lcd.print("Drive mode: ");
  lcd.setCursor(4,1);
  if (drivemode==0){
    lcd.print("Direct      ");
  }else{
    lcd.print("Interpolated");
  }
  lcd.setCursor(0,2);
  if (drivemode==0){
    lcd.print("(Direct transmission");
  }else{
    lcd.print("(Smooth transmission");
  }
  lcd.setCursor(0,3);
  lcd.print("of control inputs)");     
}

//LCD page for control mode adjustment
void Controlenc(){
  lcd.setCursor(0,0);
  lcd.print("Control mode:");
  lcd.setCursor(1,1);
  if (controlmode==0){
    lcd.print("Stick L: Tilt     "); //control mode zero: left stick controls tilt
  }else{
    lcd.print("Stick L: Track L  "); //control mode one: left stick controls left track
  }
  lcd.setCursor(1,2);
  if (controlmode==0){
    lcd.print("Stick R: Track L+R"); //control mode zero: right stick controls both tracks
  }else{
    lcd.print("Stick R: Track R  "); //control mode one: right stick controls right track
  }
  lcd.setCursor(0,3);
  if (controlmode==0){
    lcd.print("(");
    lcd.write(1); 
    lcd.print(") Use F-B and L-R");     
  }else{
    lcd.print("(");
    lcd.write(5);
    lcd.print(") Use F-B only   ");
  }
}

//LCD screen for speed adjustment
void Speedenc(){
  lcd.setCursor(0,0);
  lcd.print("Speed adjustment:");
  lcd.setCursor(1,1);
  lcd.print("Fast:");
  lcd.setCursor(7,1);
  lcd.print(speedhigh);
  if(speedhigh<100&&speedhigh>9){ //needed to get rid of old characters if the speed setting is below 100
    lcd.setCursor(9,1);
    lcd.print(" ");
  }else if(speedhigh<10){ //never needed because the speed is always larger than 9 - left from code testing
    lcd.setCursor(8,1);
    lcd.print("  ");
  }
  lcd.setCursor(11,1);
  lcd.print("(75-200)");
  lcd.setCursor(1,2);
  lcd.print("Slow:");
  lcd.setCursor(7,2);
  lcd.print(speedlow);
  if(speedlow<100&&speedlow>9){
    lcd.setCursor(9,2);
    lcd.print(" ");
  }else if(speedlow<10){
    lcd.setCursor(8,2);
    lcd.print("  ");
  }
  lcd.setCursor(11,2);
  lcd.print("(50-175)");
  lcd.setCursor(1,3);
  lcd.print("Back");
  if (adj==0){ //printing cursor position
    for (int i=0; i<3; i++){
      lcd.setCursor(0,i+1);
      if (rot==i){
        lcd.write(1);
      }else{
        lcd.print(" ");
      }
      lcd.setCursor(6,1);
      lcd.print(" ");
      lcd.setCursor(6,2);
      lcd.print(" ");
    }
  }else{ //printing cursor position when entering one of the speed adjust options
    if (adj==1){
      lcd.setCursor(6,1);
      lcd.write(1);
      lcd.setCursor(6,2);
      lcd.print(" ");
    }else{
      lcd.setCursor(6,2);
      lcd.write(1);
      lcd.setCursor(6,1);
      lcd.print(" "); 
    }
  }
}

//updating the RGB code for the data package that is sent to the tank
void RGBupdate() {
  switch (program) {
    case 0: //program zero: running light off
      outgoing.R = 0;
      outgoing.G = 0;
      outgoing.B = 0;
      break;
    case 1: //program one: running light blue
      outgoing.R = 0;
      outgoing.G = 0;
      outgoing.B = 127;
      break;
    case 2: //program two: running light cyan
      outgoing.R = 0;
      outgoing.G = 127;
      outgoing.B = 127;
      break;
    case 3: //program three: running light green
      outgoing.R = 0;
      outgoing.G = 127;
      outgoing.B = 0;
      break;
    case 4: //program four: running light yellow
      outgoing.R = 127;
      outgoing.G = 127;
      outgoing.B = 0;
      break;
    case 5: //program five: running light red
      outgoing.R = 127;
      outgoing.G = 0;
      outgoing.B = 0;
      break;
    case 6: //program six: running light violet
      outgoing.R = 127;
      outgoing.G = 0;
      outgoing.B = 127;
      break;
    case 7: //program seven: running light white
      outgoing.R = 127;
      outgoing.G = 127;
      outgoing.B = 127;
      break;
    case 8: //program eight: running light rainbow
      outgoing.R = rbr;
      outgoing.G = rbg;
      outgoing.B = rbb;
      break;          
  }  
}

void speedchk1 (){ //movement control with two joysticks
  xval = analogRead(xaxis); //front-back, right   
  if (xval>3780){
    xval = 3780;  //individual correction to center this joystick-ADC combination (no correction for the non-linearity, only to set the center position to zero) -> individual for every ESP32
  }
  xval = map(xval,0,3780,-speedmax,speedmax); //mapping of this value between +-maximum speed
  if (abs(xval)<(speedmax/12)){ //very low speeds are set to zero to avoid false inputs due to sensor noise (these low values are anyway below the threshold for actual moving (duty cycle too low)
    xval = 0;
  }  
  outgoing.spdR = xval;  
  
  xaxis2v = analogRead(xaxis2); //front-back, left  
  if (xaxis2v>3600){
    xaxis2v = 3600;  //individual correction to center this joystick-ADC combination
  }  
  xaxis2v = map(xaxis2v,0,3600,-speedmax,speedmax);
  if (abs(xaxis2v)<(speedmax/12)){
    xaxis2v = 0;
  }
  outgoing.spdL = xaxis2v;
}

void speedchk0 (){ //movement control with one joystick
  yval = analogRead(yaxis); //left-right, right
  if (yval>3680){
    yval = 3680;  //individual correction to center this joystick-ADC combination
  }
  yval = map(yval,0,3680,-speedmax,speedmax);
  if (abs(yval)<(speedmax/12)){
    yval = 0;
  }
 
  xval = analogRead(xaxis); //front-back, right
  if (xval>3780){
    xval = 3780;  //individual correction to center this joystick-ADC combination
  }
  xval = map(xval,0,3780,-speedmax,speedmax);
  if (abs(xval)<(speedmax/12)){
    xval = 0;
  }
    
  if (xval<0){ //specifying movement direction (forward & reverse) based on front-back position of the right joystick
    dir = 0;
  }else{
    dir = 1;
  }
  
  spdL = abs(xval); //specifying the global speed for both tracks...
  spdR = abs(xval);
  if (yval<0){ //...before selectively reducing the speed of one side of the tracks depending on the left-right position of the right joystick...
    if (spdL+yval>0){
      spdL = spdL+yval;
    }else{
      spdL = 0;
    }
  }
  
  if (yval>0){ //...the same for the other track...
    if (spdR-yval>0){
      spdR = spdR-yval;
    }else{
      spdR = 0;
    }
  }
  
  if (dir==0){ //...and finally adding the direction to the output for the tank
    outgoing.spdL = -spdL;
    outgoing.spdR = -spdR;
  }else{
    outgoing.spdL = spdL;
    outgoing.spdR = spdR;
  }
}

//LCD screen for adjusting the headlight intensity
void Lightenc(){
  lcd.setCursor(0,0);
  lcd.print("Light intensity:   ");
  lcd.setCursor(0,1);
  lcd.print("    ");
  lcd.setCursor(4,1);
  lcd.print(ledv);
  lcd.setCursor(0,3);
  lcd.print("(Range: 1 - 4)");
}

//LCD screen for the selection menu
void LCDenc(){
  for (int i=0; i<8; i++){ //cursor position
    if (i>3){ //second row
      lcd.setCursor(10,i-4);
      if (i==rot){
        lcd.write(1);
      }else{
        lcd.print(" ");
      }           
    }else{ //first row
      lcd.setCursor(0,i);
      if (i==rot){
        lcd.write(1);
      }else{
        lcd.print(" ");
      }      
    }
  }

  lcd.setCursor(1,0);
  lcd.print("Back    ");
  lcd.setCursor(1,1);
  lcd.print("LED int.");
  lcd.setCursor(1,2);
  lcd.print("Speed   ");
  lcd.setCursor(1,3);
  lcd.print("Control ");
  lcd.setCursor(11,0);
  lcd.print("Drive   ");
  lcd.setCursor(11,1);
  lcd.print("Tilt    ");
  lcd.setCursor(11,2);
  lcd.print("RGB mode");
  lcd.setCursor(11,3);
  lcd.print("Weapon  ");  
}

//LCD default screen update
void updateLCD(){
  lcd.setCursor(4,0);
  lcd.print("Spd");
  if (speedmax>159){ //max speed indicator using the (partially) filled bar character that was defined in the beginning
    lcd.write(5);
  }else if (speedmax>119){
    lcd.write(4);
  }else if (speedmax>79){
    lcd.write(3);
  }else{
    lcd.write(2);
  }
  if (rswv==1){ //indication of fast or slow speed mode
    lcd.print("F");
  }else{
    lcd.print("S");
  }
  if (drivemode==0){ //indication of direct or interpolated drive mode
    lcd.print("D");
  }else{
    lcd.print("I");
  }
  lcd.setCursor(11,0);
  lcd.print("Hgt");
  if (potival>44){ //height setting displayed by (partially) filled bar
    lcd.write(2);
  }else if (potival>34){
    lcd.write(3);
  }else if (potival>24){
    lcd.write(4);
  }else{
    lcd.write(5);
  }
  if (defaultheight==110){ //indication of automatic or manual tilt adjust mode
    lcd.print("A");  
  }else if (defaultheight==111){
    lcd.print("M");
  }

  if (txbatwarn==1||rxbatwarn==1){ //the voltages of receiver (tank) and transmitter (remote control) blink when the battery low warning (not battery low error!) is active
    batwarn = batwarn+1;
  }  
  lcd.setCursor(6,1);
  lcd.print("RX ");
  if (dnsv==1){ //"do not show voltage" indicates whether voltage should be shown (and it should not because the "X" is shown when the wireless connection is lost
    lcd.write(6);
    lcd.print("    ");
  }else if (batwarn<8||rxbatwarn==0){ //show the voltage when there is no battery warning or when the counter is low
    lcd.print(voltage, 1);
    lcd.print("V");
  }else if (batwarn>7&&rxbatwarn==1){ //do not show the voltage when there is battery warning and the counter is high => results in blinking for the warning
    lcd.print("     ");
  }
  lcd.setCursor(6,2);
  lcd.print("TX "); //remote control voltage is always shown because there cannot be a connection loss
  if (batwarn<8||txbatwarn==0){
    lcd.print(RCvol, 1);
    lcd.print("V");    
  }else if(batwarn>7&&txbatwarn==1){
    lcd.print("    ");
  }
  if (batwarn>9){
    batwarn = 0;
  }

  lcd.setCursor(0,3);
  lcd.print("LED");
  lcd.setCursor(3,3);
  if (lighton==1){ //the LED headlight status: X stands for off; the (partially) filled bar indicates intensity when active
    switch (ledv){
      case 4:
        lcd.write(5);
        break;
      case 3:
        lcd.write(4);
        break;
      case 2: 
        lcd.write(3);
        break;
      case 1:
        lcd.write(2);
        break;     
    }
  }else{
    lcd.write(6);
  }

  if (controlmode==0){ //control mode zero means left joystick for tilt adjustment
    if (defaultheight==110){ //defaultheight 110 means auto tilt adjustment
      if(outgoing.hv>100&&outgoing.hv<110){ //indicate with slash that tilt modification is not possible with this setting when attempting to adjust tilt
        for (int i=0; i<3; i++){
          for (int j=0; j<3; j++){
            lcd.setCursor(i,j);
            if ((i==0&&j==0)||(i==2&&j==0)||(i==1&&j==1)||(i==0&&j==2)||(i==2&&j==2)){
              lcd.print("/");
            }else{
              lcd.print(" ");
            }
          }
        }             
      }else{
        for (int i=0; i<3; i++){ //show a slash in the center of the joystick indicator when using this setting combination
          for (int j=0; j<3; j++){
            lcd.setCursor(i,j);
            if (i==1&&j==1){
              lcd.print("/");
            }else{
              lcd.print(" ");
            }
          }
        }          
      }

    }else{ //defaultheight 111 means manual tilt adjustment
      for (int i=0; i<3; i++){  //Display the position of the left joystick (indicating if tilt adjustment is active)
        for (int j=0; j<3; j++){
          lcd.setCursor(j,i);
          if (outgoing.hv<101&&j==1&&i==1){
            lcd.write(1); 
          }else if (outgoing.hv==101&&j==0&&i==1){
            lcd.write(1); 
          }else if (outgoing.hv==102&&j==0&&i==0){
            lcd.write(1); 
          }else if (outgoing.hv==103&&j==1&&i==0){
            lcd.write(1); 
          }else if (outgoing.hv==104&&j==2&&i==0){
            lcd.write(1); 
          }else if (outgoing.hv==105&&j==2&&i==1){
            lcd.write(1); 
          }else if (outgoing.hv==106&&j==2&&i==2){
            lcd.write(1); 
          }else if (outgoing.hv==107&&j==1&&i==2){
            lcd.write(1); 
          }else if (outgoing.hv==108&&j==0&&i==2){
            lcd.write(1); 
          }else if (outgoing.hv==109){
            if ((j==1&&i==0)||(j==0&&i==1)||(j==2&&i==1)||(j==1&&i==2)){
              lcd.write(1);
            }else{
              lcd.print(" ");
            }
          }else if(j==1&&i==1){
            lcd.write(1);
          }else{
            lcd.print(" ");
          }
        }
      }
    }
    for (int i=0; i<3; i++){  //Display the position of the right joystick (still in controlmode=0: right joystick is responsible for driving both tracks)
      for (int j=17; j<20; j++){
        lcd.setCursor(j,i);
        if (outgoing.spdR>0&&outgoing.spdL==0&&j==17&&i==0){
          lcd.write(1); 
        }else if (outgoing.spdL>0&&outgoing.spdR>0&&j==18&&i==0){
          lcd.write(1); 
        }else if (outgoing.spdL>0&&outgoing.spdR==0&&j==19&&i==0){
          lcd.write(1); 
        }else if (outgoing.spdR<0&&outgoing.spdL==0&&j==17&&i==2){
          lcd.write(1); 
        }else if (outgoing.spdL<0&&outgoing.spdR<0&&j==18&&i==2){
          lcd.write(1); 
        }else if (outgoing.spdR==0&&outgoing.spdL<0&&j==19&&i==2){
          lcd.write(1); 
        }else if(j==18&&i==1){
          lcd.write(1);
        }else{
          lcd.print(" ");
        }
      }
    }      
  }else{ //controlmode=1: left joystick controls left track; right joystick controls right track
    for (int i=0; i<3; i++){ //Display the position of the right joystick (only front-back is relevant)
      lcd.setCursor(18,i);
      if(outgoing.spdR>0&&i==0){
        lcd.write(5); //note the different symbol for this control mode!
      }else if(outgoing.spdR<0&&i==2){
        lcd.write(5);
      }else if(i==1){
        lcd.write(5);
      }else{
        lcd.print(" ");
      }
    }
    for (int i=0; i<3; i++){ //Display the position of the left joystick (only front-back is relevant)
      lcd.setCursor(1,i);
      if(outgoing.spdL>0&&i==0){
        lcd.write(5);
      }else if(outgoing.spdL<0&&i==2){
        lcd.write(5);
      }else if(i==1){
        lcd.write(5);
      }else{
        lcd.print(" ");
      }
    }
  }

  lcd.setCursor(5,3);
  lcd.print("Cxn");
    lcd.setCursor(8,3);
    dnsv = 0; //"do not show voltage" is zero by default
    if (signalquality>16){ //display the connection quality using the (partially) filled bar symbol
      lcd.write(5);
    }else if (signalquality>13){
      lcd.write(4);
    }else if (signalquality>10){
      lcd.write(3);
    }else if (signalquality>7){
      lcd.write(2);
    }else{
      lcd.write(6);
      dnsv = 1; //it is set to one if the signal is lost
    }

  lcd.setCursor(10,3);
  lcd.print("Wpn");
  if(lswv==1){
    switch (force){
      case 4:
        lcd.write(5); //the force of the weapon is indicated by the (partially) filled bar symbol if it is active...
        break;
      case 3:
        lcd.write(4);
        break;
      case 2: 
        lcd.write(3);
        break;
      case 1:
        lcd.write(2);
        break;     
    }
  }else{
    lcd.write(6); //...and the X if it is disabled
  }
  if(bullets==17){
    lcd.print("XX"); //sign for jammed weapon
  }else if(bullets>9){
    lcd.print(bullets); //printing the number of remaining bullets
  }else{
    lcd.print("0");
    lcd.print(bullets);
  }
  
  lcd.setCursor(17,3);
  switch (program){ //printing the program of the RGB LED strip using three-character codes for the colors
    case 0:
      lcd.print("Off");
      break;
    case 1:
      lcd.print("Blu");
      break;
    case 2:
      lcd.print("Cyn");
      break;
    case 3:
      lcd.print("Grn");
      break;
    case 4:
      lcd.print("Yel");
      break;
    case 5:
      lcd.print("Red");
      break;
    case 6:
      lcd.print("Vio");
      break;
    case 7:
      lcd.print("Wht");
      break;
    case 8:
      lcd.print("Rbw");
      break;
  }  
}

//the height adjust function reads the potentiometer of the remote control that governs the main height of the tank body and the inputs for manual tilt adjustment
void heightAdjust(){
  potival = analogRead(poti);
  for (int i=0; i<19; i++){ //noise reduction by averaging 20 readings of the ADC
    potival = potival+analogRead(poti);
  }
  potival = potival/20;
  if (rswv>0){
    potival = map(potival,0,4095,40,55);
  }else{
    potival = map(potival,0,4095,15,55);
  }
  outgoing.hv = potival;

  if (controlmode==0){ //manual tilt adjustment
    yaxis2v = analogRead(yaxis2); //left joystick
    yaxis2v = map(yaxis2v,0,4095,-127,127); //no center calibration here because it is not needed (tilt adjustment is basically digital but derived from the analog signal)
  
    xaxis2v = analogRead(xaxis2);
    xaxis2v = map(xaxis2v,0,4095,-127,127);
    
    if (yaxis2v<-100&&abs(xaxis2v)<100){ //threshold for sending tilt adjustment commands: 100 of 127
      outgoing.hv = 101;
    }else if (yaxis2v<-100&&xaxis2v>100){ //the combinations of front, front-right, right, back-right, back, back-left, left, and front-left are converted to single numbers for transmission
      outgoing.hv = 102;
    }else if (abs(yaxis2v)<100&&xaxis2v>100){
      outgoing.hv = 103;
    }else if (yaxis2v>100&&xaxis2v>100){
      outgoing.hv = 104;
    }else if (yaxis2v>100&&abs(xaxis2v)<100){
      outgoing.hv = 105;
    }else if (yaxis2v>100&&xaxis2v<-100){
      outgoing.hv = 106;
    }else if (abs(yaxis2v)<100&&xaxis2v<-100){
      outgoing.hv = 107;
    }else if (yaxis2v<-100&&xaxis2v<-100){
      outgoing.hv = 108;
    }
  }    
}

//the serial monitor was used for debugging
void updateSerial(){
//  Serial.print("Height adjust mode:  ");
//  if (defaultheight==110){
//    Serial.println("Auto");
//  }else if (defaultheight==111){
//    Serial.println("Manual");
//  }else{
//    Serial.println("Error");
//  }
//  Serial.print("Left switch: ");
//  Serial.print(lswv);
//  Serial.print(", right switch: ");
//  Serial.println(rswv);
//  Serial.print("Left button: ");
//  Serial.print(lbtnv);
//  Serial.print(", right button: ");
//  Serial.println(rbtnv);
//  Serial.print("Poti: ");
//  Serial.print(potival);
//  Serial.print(", Encoder: ");
//  Serial.println(encbtnv);
//  Serial.print("RC voltage: ");
//  Serial.println(RCvol);
//  Serial.print("Crawler battery status: ");
//  Serial.print(voltage, 2);
//  Serial.println(" V");
//  Serial.print("Speed values: ");
//  Serial.print(outgoing.spdL);
//  Serial.print(", ");
//  Serial.println(outgoing.spdR);
//  Serial.print("Data for Serial2: ");
//  Serial.println(outgoing.hv);  
//  Serial.print("Checking encoder? ");
//  if (encchk==0){
//    Serial.println("Y");
//  }else{
//    Serial.println("N");
//  }
//  Serial.print("Encoder position: ");
//  Serial.println(rot);
//  Serial.println("----------");
}
